# Copyright (C) 2015 by Dalen Bernaca
# YOU ARE ALLOWED TO USE AND DISTRIBUTE THIS PROGRAM FREELY
# AS THIS PROGRAM IS BOUND BY MBROLA LICENCE AS WELL, USING IT FOR COMMERCIAL OR MILITARY PURPOSES IS STRICTLY FORBIDDEN
#
# YOU ARE ALLOWED TO MODIFY THIS PROGRAM IN ONE CONDITION:
# THAT YOU SEND YOUR MODIFIED VERSION TO THE AUTHOR BY E-MAIL
# WITH SHORT EXPLANATION REGARDING THE MODIFICATION(S)
# IF YOUR MODIFICATION(S) INCREASE THE QUALITY OF THE PROGRAM
# THEY WILL BE TAKEN IN VIEW FOR THE NEW OFFICIAL VERSION
# THIS PARAGRAPH EXCLUDES THE MBROLA PART OF THIS SOFTWARE, OF COURSE
#
# NOTE THE MBROLA LICENCE WHICH WILL BE PRESENTED TO YOU WHEN USING THIS SOFTWARE
# AND ANY PARTICULAR LICENCE DISTRIBUTED WITH EACH MBROLA VOICE DIPHONE DATABASE
#
# DIFFERENT LICENCES MAY BE USED BY AUTHORS OF PRONUNCIATION (*.mbx) SCRIPTS DISTRIBUTED ALONG WITH MBROLA DIPHONE DATABASES
# AS PART OF THE MBROLAX VOICE
# TO SEE THEM YOU SHOULD EXAMINE THE SOURCE CODE OF (*.mbx) SCRIPT YOU ARE INTERESTED IN
#
# AUTHOR OF THIS SOFTWARE IS RESPECTING ALL THESE LICENCES AND SO SHOULD YOU
#
# Author:   Dalen Bernaca
#           dbernaca@blue-merlin.hr
# Version:  1.0B
# Homepage: http://www.brailleweb.com/Mbrolax/
#
# THE AUTHOR IS NOT RESPONSIBLE FOR ANY DAMAGE INFLICTED BY MISUSE/ABUSE OF THIS SOFTWARE
# ANY OTHER PROGRAM WITH SAME/SIMILAR NAME AND PURPOSE WHICH IS NOT PROCURED FROM
# http://www.brailleweb.com/,
# http://www.sciencepower.org/ or
# http://apaslab.riteh.hr
# IS NOT THE OFFICIAL VERSION WHATEVER MAY BE WRITTEN IN THE SOURCE CODE AND
# SHOULD BE TREATED WITH ENORMOUS CARE
# THE AUTHOR IS NOT ACCOUNTABLE FOR SUCH SOURCE CODE

"""
This is an add-on for NVDA screen reader that allows NVDA to use Mbrolax TTS engine directly.
Mbrolax uses mbrola synthesizer to synthesize speech.
It prepares text in a given language to be synthesized with a given mbrola diphone database.
In the end, it passes this input to mbrola in PHO format and reproduces the returned audio stream
with its own audio output methods. In this case audio output relies on NVDA nvwave module.

Mbrolax supports contextual natural language processing with simple interpreting language in which you can specify preprocessing rules
which will be applied on input text.

Also, it is easily extendable.
Anyone can add new voice to Mbrolax using the mbrola diphone database and Python programming language to explain to Mbrolax how to convert text to PHO string to be used with that database.

Anyone who wishes to improve the Mbrolax engine with stuff that does not include voices should contact the author even before doing so because
Mbrolax internal code is very fluid and will not keep same interfaces throughout different versions.
Backward compatibility is not guaranteed.

As for *.mbx voice scripts, the outer public interface also may not stay the same, but each official *.mbx script will be accordingly changed by the author.
So, incompatibilities will never happen if users stick to official *.mbx scripts. Which is what they should do primarily for security reasons.
*.mbx pronunciation scripts have to be sent to the author to be checked, approved and published on official sites.
Documentation will also be updated to match any outer interface changes.
It can be found on official Mbrolax web page:
http://brailleweb.com/Mbrolax/

Many thanks to Olga Yakovleva, Aleksey Sadovoy and Michael Curran whose NVDA add-ons for Festival and SVOXPico I used as examples on how to write a NVDA add-on.
"""

from synthDriverHandler import SynthDriver, VoiceInfo
from mbxvoice           import VoiceError
from logHandler         import log
from collections        import OrderedDict, deque
from time import        sleep
import synth as mbx
import speech
import nvwave
import config
import threading
import os

__author__  = "Dalen Bernaca"
__version__ = "1.0B"

MbrolaxPath = os.path.dirname(os.path.abspath(__file__))
# Py2Exe support:
if os.path.isfile(MbrolaxPath): MbrolaxPath = os.path.dirname(MbrolaxPath)


class SynthDriver(SynthDriver):
    name = "Mbrolax"
    description = "Mbrolax TTS engine"
    supportedSettings = (SynthDriver.VoiceSetting(), SynthDriver.RateSetting(), SynthDriver.PitchSetting(), SynthDriver.InflectionSetting(), SynthDriver.VolumeSetting())
    availableVoices = OrderedDict()
    for x in mbx.voices():
        vn, vd, vl = x[1:4]
        availableVoices[vn] = VoiceInfo(vn, vd, vl)
    _voice = "en1"
    pitch  = 50
    rate   = 50

    @classmethod
    def check(cls):
        return len(cls.availableVoices)!=0

    def __init__ (self):
        self._voice = self.availableVoices.keys()[0]
        for x in self.availableVoices.keys():
            if x.lower().startswith("en"): self._voice = x; break
        mbx.init(self._voice)
        self.rate       = self._paramToPercent(mbx.curvoice.defaults["speed"], 20, 300)
        self.pitch      = self._paramToPercent(mbx.curvoice.defaults["pitch"], 70, 600)
        self.pitchscale = mbx.curvoice.defaults["pitchscale"]
        self.flatscale  = mbx.curvoice.defaults["flatscale"]
        self.wordpause  = mbx.curvoice.defaults["wordpause"]
        self.punctpause = mbx.curvoice.defaults["punctpause"]
        self.queue = deque()
        self.player = nvwave.WavePlayer(channels=1, samplesPerSec=mbx.mbrola.getFreq(), bitsPerSample=16, outputDevice=config.conf["speech"]["outputDevice"])
        self.isSpeaking = False
        self.background_thread = threading.Thread(target=self.background_thread_func)
        self.background_thread.daemon = True
        self.background_thread.start()
        log.info("Mbrolax %s using Mbrola DLL %s" % (__version__, mbx.mbrola.version()))

    def terminate (self):
        self.cancel()
        self.queue.append((None, None, None))
        self.background_thread.join()
        self.player.close()
        self.player = None
        mbx.deinit()

    def _get_voice(self):
        return self._voice

    def _set_voice (self, value):
        value = str(value)
        if value in self.availableVoices:
            self.cancel()
            self._voice = value; mbx.init(self._voice)
            self.rate  = self._paramToPercent(mbx.curvoice.defaults["speed"], 20, 300)
            self.pitch      = self._paramToPercent(mbx.curvoice.defaults["pitch"], 70, 600)
            self.pitchscale = mbx.curvoice.defaults["pitchscale"]
            self.flatscale  = mbx.curvoice.defaults["flatscale"]
            self.wordpause  = mbx.curvoice.defaults["wordpause"]
            self.punctpause = mbx.curvoice.defaults["punctpause"]
        else: raise VoiceError, "Voice '%s' not supported" % value

    def getnext (self):
        decay = 0
        while True:
            try: return self.queue.popleft()
            except:
                sleep(0.002+decay)
                if decay<=0.18: decay += 0.000008

    def background_thread_func(self):
        while True:
            text, index, pitch = self.getnext()
            self.lastIndex = index
            if text is None: break
            self.isSpeaking = True
            for wordaudio in mbx.itersynth(text, self._percentToParam(self.rate, 20, 300),
                                self._percentToParam(pitch, 70, 600),
                                self.pitchscale, self.flatscale, self.wordpause, self.punctpause):
                if not self.isSpeaking: break
                self.player.feed(wordaudio)
            self.isSpeaking = False

    def cancel(self):
        self.queue.clear()
        self.isSpeaking = False
        self.player.stop()
        self.lastIndex = None

    def speak(self, speechSequence):
        self.cancel()
        textList = []
        index = None
        for item in speechSequence:
            if isinstance(item, basestring):
                textList.append(item)
            elif isinstance(item, speech.IndexCommand):
                index = item.index
        text = " ".join(textList)
        if text:
            self.queue.append((text, index, self.pitch))

    def pause(self, switch):
        self.player.pause(switch)

    def _get_inflection(self):
        return self._paramToPercent(self.pitchscale, 1, 30)

    def _set_inflection (self, value):
        self.pitchscale = self._percentToParam(value, 1, 30)

    def _set_volume (self, v):
        v = v/100.0
        mbx.mbrola.setVolume(v)

    def _get_volume (self):
        return int(mbx.mbrola.getVolume()*100)
